'''trec_dd.harness.run provides an evaluation jig for TREC Dynamic Domain systems

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.
'''

from __future__ import absolute_import, print_function

import argparse
from collections import defaultdict
from dossier.label import LabelStore, Label, CorefValue
import json
import itertools
import kvlayer
import logging
import os
import sys
import time
import yakonfig

from trec_dd.harness.truth_data import parse_truth_data

logger = logging.getLogger(__name__)


def query_to_topic_id(query):
    return query.replace(' ', '_')


def topic_id_to_query(topic_id):
    return topic_id.replace('_', ' ')


TOPIC_IDS = 'trec_dd_harness_topic_ids'
EXPECTING_STOP = 'trec_dd_harness_expecting_stop'

class Harness(object):

    tables = {
        TOPIC_IDS: (str,),
        EXPECTING_STOP: (str,),
    }

    def __init__(self, config, kvl, label_store):
        self.kvl = kvl
        self.kvl.setup_namespace(self.tables)
        self.label_store = label_store
        self.truth_data_path = config.get('truth_data_path')
        self.run_file_path = config.get('run_file_path')
        self.topic_ids = set(config.get('topic_ids', []))
        self.batch_size = int(config.get('batch_size', 5))

    config_name = 'harness'

    def verify_label_store(self):
        ls = iter(self.label_store.everything())
        try:
            ls.next()
        except StopIteration:
            sys.exit('The label store is empty.  Have you run `trec_dd_harness load`?')
        else:
            return True

    def init(self, topic_ids=None):
        '''Initialize the DB table of topics to apply to the engine under test.
        '''
        self.kvl.clear_table(TOPIC_IDS)
        all_topics = dict()
        for label in self.label_store.everything():
            all_topics[(label.meta['topic_id'],)] = label.meta['topic_name']
        # allow in-process caller to init with topic ids of its choosing
        if topic_ids is not None:
            self.topic_ids = set(topic_ids)
        if self.topic_ids:
            for topic_id in all_topics.keys():
                if topic_id not in self.topic_ids:
                    all_topics.pop(topic_id)
        self.kvl.put(TOPIC_IDS, *all_topics.items())
        return {'num_topics': len(all_topics)}

    def check_expecting_stop(self):
        for (topic_id,), _ in self.kvl.scan(EXPECTING_STOP):
            sys.exit('Harness was expecting you to call stop because you '
                     'submitted fewer than batch_size results.  Fix your '
                     'system and try again.')

    def unset_expecting_stop(self):
        self.kvl.clear_table(EXPECTING_STOP)

    def set_expecting_stop(self, topic_id):
        self.kvl.put(EXPECTING_STOP, ((topic_id,), 'YES'))

    def start(self):
        '''initiates a round of feedback to recommender under evaluation.
        '''
        self.check_expecting_stop()
        self.verify_label_store()
        for (topic_id,), query_string in self.kvl.scan(TOPIC_IDS):
            return {'topic_id': topic_id, 'query': query_string}

        # finished all the topics, so end.
        return {'topic_id': None, 'query': None}

    def stop(self, topic_id):
        '''ends a round of feedback
        '''
        self.unset_expecting_stop()
        for idx, ((_topic_id,), query_string) in enumerate(self.kvl.scan(TOPIC_IDS)):
            if idx == 0:
                if topic_id != _topic_id:
                    sys.exit('%d != %d, which is where the database says we are'
                             % (topic_id, _topic_id))
                self.kvl.delete(TOPIC_IDS, (topic_id,))
                logger.info("Finished with topic: '%s'", topic_id)
        return {'finished': topic_id, 'num_remaining': idx }

    def step(self, topic_id, results):
        '''Generates feedback on one round of recommendations
        '''
        self.check_expecting_stop()
        self.verify_label_store()
        query_string = None
        for (_topic_id,), query_string in self.kvl.scan(TOPIC_IDS):
            break
        if query_string is None:
            sys.exit('got out of sync: topic_id=%r' % topic_id)
        if topic_id != _topic_id:
            sys.exit('%d != %d, which is where the database says we are'
                     % (topic_id, _topic_id))

        if len(results) > 2 * self.batch_size:
            logger.warn('command="step" allows up to twice batch_size (2 x %d = %d) '
                        'input args: stream_id conf stream_id conf ..., not %d = len(%r)',
                        self.batch_size, 2 * self.batch_size, len(results), results)
            return {'error': 'MUST EXIT: submitted too many results'}

        if len(results) < 2 * self.batch_size:
            logger.warn('fewer than the batch size, so automatically calling `stop` '
                        'this query; you must call `start` to move on to the next query.')
            self.set_expecting_stop(topic_id)
    
        pairs = [iter(results)] * 2
        results = [(stream_id, int(conf))
                   for stream_id, conf in itertools.izip_longest(*pairs)]

        # private function for constructing feedback, used in `map` below
        def feedback_for_result(result):            
            stream_id, confidence = result
            if len(stream_id.strip()) == 0:
                sys.exit('Your system submitted a bogus document identifier: %r'
                         % stream_id)
            try:
                assert 0 <= int(confidence) <= 1000
            except:
                sys.exit('Your system submitted a bogus confidence value: %r'
                         % confidence)

            labels_for_doc = self.label_store.directly_connected(stream_id)
            labels_for_doc = filter(lambda l: l.other(stream_id) == topic_id,
                                    labels_for_doc)

            # If any of the labels between the topic_id and
            # the document are negative, we call this document
            # off-topic. If there are no labels between this
            # topic_id and the document, we call this document
            # off-topic. Otherwise, we extract the subtopics
            # from the labels and call the document on-topic.
            if any([label.value == CorefValue.Negative
                    for label in labels_for_doc]):
                subtopic_feedback = []
            else:
                def subtopic_from_label(label):
                    subtopic_id = label.subtopic_for(stream_id)
                    offset_begin, offset_end = map(int, subtopic_id.split(','))
                    subtopic = {
                        'subtopic_id': label.subtopic_for(topic_id),
                        'subtopic_name': label.meta['subtopic_name'],
                        'offset_begin': offset_begin,
                        'offset_end': offset_end,
                        'passage_text': label.meta['passage_text'],
                        'rating': label.rating,
                    }
                    return subtopic
                
                subtopic_feedback = [subtopic_from_label(label)
                                     for label in labels_for_doc]

                #subtopic_feedback = []
                #for _, data in subtopic_id_to_data.iteritems():
                #    best = max(data, key=lambda d: d['rating'])
                #    subtopic_feedback.append(best)

            feedback = {
                'topic_id': topic_id,
                'confidence': confidence,
                'stream_id': stream_id,
                'subtopics': subtopic_feedback,
                'on_topic': int(bool(len(subtopic_feedback) > 0))
            }

            return feedback

        all_feedback = map(feedback_for_result, results)
        self.write_feedback_to_run_file(all_feedback)
        return all_feedback


    def write_feedback_to_run_file(self, feedback):
        if self.run_file_path is None:
            return

        # *append* to the run file
        run_file = open(self.run_file_path, 'a')

        for entry in feedback:
            subtopic_stanza = 'NULL'
            if entry['subtopics']:
                subtopic_tuples = []
                for subtopic in entry['subtopics']:
                    subtopic_tuple = ':'.join([subtopic['subtopic_id'],
                                               str(subtopic['rating'])])
                    subtopic_tuples.append(subtopic_tuple)
                subtopic_stanza = '|'.join(subtopic_tuples)

            # <topic> <document-id> <confidence> <on_topic> <subtopic data>
            run_file_line = ('%(topic_id)s\t%(stream_id)s\t%(confidence).6f'
                             '\t%(on_topic)d\t%(subtopic_stanza)s\n')
            output_dict = {'subtopic_stanza': subtopic_stanza,
                           'topic_id': entry['topic_id'],
                           'stream_id': entry['stream_id'],
                           'confidence': entry['confidence'],
                           'on_topic': entry['on_topic'],
                           }
            to_write = run_file_line % output_dict

            assert len(run_file_line.split()) == 5

            run_file.write(to_write)

        run_file.close()

usage = '''The purpose of this harness is to interact with your TREC DD system
by issuing queries to your system, and providing feedback (truth data)
for the results produced by your system.  While it does this, it keeps
track of the results produced by your system in a `run file`.  After
generating a run file with this harness, you can score the run using
trec_dd_scorer

The harness is run via  three commands: start, step, stop.  Typically,
a system will invoke start, then  invoke step multiple times, and then
invoke stop.   Every invocation  must include the  -c argument  with a
path   to    a   valid    config.yaml   file,   as    illustrated   in
example/config.yaml.  For  efficiency, the first  time you run  with a
new configuration,  the truth data  must be loaded into  your database
using the `load` command.  

By default, when you score a system using the harness, all of the
topics are applied to the system in an order selected by the harness.
You can limit the topic_ids that are used by specifying the topic_ids
property in the config.yaml

The harness keeps track of the topic_ids that have not yet been used
in building your system's run file.  To reset this state, you must run
the `init` command.

To progress through the topics, your system must execute this double
while loop, which is exactly what is implemented in the
trec_dd/system/ambassador_cli.py example:

        `init`
        while 1:
            topic_id <-- `start`
            if topic_id is None: break
            while 1:
                results <-- run your system
                feedback <-- `step(results)`
                if feedback is None or len(feedback) < batch_size:
                    break
                else:
                    your system processes the feedback
            `stop`

Each of the five commands returns a JSON dictionary which your system
can read using a JSON library.  The harness always provides feedback
for every result, even if the feedback is that the system has no truth
data for that result.  Note that your use of the harness *must* call
`stop` in the next iteration after any step in which you submit fewer
than batch_size results.  If you fail to do this, the harness will
exit.

See trec_dd/system/ambassador_cli.py for an example of using the
harness from python.

'''

def main():
    parser = argparse.ArgumentParser(
        'Command line interface to the office TREC DD jig.',
        usage=usage,
        conflict_handler='resolve')
    parser.add_argument('command', help='must be "load", "init", "start", "step", or "stop"')
    parser.add_argument('args', help='input for given command',
                        nargs=argparse.REMAINDER)
    modules = [yakonfig, kvlayer, Harness]
    args = yakonfig.parse_args(parser, modules)

    logging.basicConfig(level=logging.DEBUG)

    if args.command not in set(['load', 'init', 'start', 'step', 'stop']):
        sys.exit('The only valid commands are "load", "init", "start", "step", and "stop".')

    kvl = kvlayer.client()
    label_store = LabelStore(kvl)
    config = yakonfig.get_global_config('harness')
    harness = Harness(config, kvl, label_store)

    if args.command == 'load':
        if not config.get('truth_data_path'):
            sys.exit('Must provide --truth-data-path as an argument')
        if not os.path.exists(config['truth_data_path']):
            sys.exit('%r does not exist' % config['truth_data_path'])
        parse_truth_data(label_store, config['truth_data_path'])
        logger.info('Done!  The truth data was loaded into this '
                     'kvlayer backend:\n%s',
                     json.dumps(yakonfig.get_global_config('kvlayer'), 
                                indent=4, sort_keys=True))

    elif args.command == 'init':
        response = harness.init()
        print(json.dumps(response))

    elif args.command == 'start':
        response = harness.start()
        print(json.dumps(response))

    elif args.command == 'stop':
        response = harness.stop(args.args[0])
        print(json.dumps(response))

    elif args.command == 'step':
        parts = args.args
        topic_id = parts.pop(0)
        feedback = harness.step(topic_id, parts)
        print(json.dumps(feedback))
            

if __name__ == '__main__':
    main()



'''trec_dd.scorer.u_err provides a u-ERR scorer for TREC DD

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''

from __future__ import division
from operator import attrgetter

#from numpy import mean

from trec_dd.utils import get_all_subtopics

def mean(l):
    if len(l) == 0:
        return 0.0

    s = sum(l)
    return s / len(l)


def u_err(run, label_store):
    scores_by_topic = dict()
    for topic_id, results in run['results'].items():
        #subtopic_ids = map(attrgetter('subtopic_id'), 
        #                   get_all_subtopics(label_store, topic_id))
        subtopic_ids = set(get_all_subtopics(label_store, topic_id))
        scores_by_topic[topic_id] = len(results) / len(subtopic_ids)

    macro_avg = mean(scores_by_topic.values())
    run['scores']['u_err'] = {'scores_by_topic': scores_by_topic, 'macro_average': macro_avg}




'''Demonstrate generating a run_file using a random system.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

This file sets up a very silly system for evaluation. All the system
does when given a query is return a collection of random document ids
from a set of known document ids.

We feed this system into a HarnessAmbassador, which orchestrates the
communication between the Harness and the system being
evaluated. Because our random system is in python, we can use the
Harness's python interface for simplicity.
'''
from __future__ import absolute_import
import argparse
from collections import defaultdict
from itertools import chain
import logging
import os
import random
import sys
import yaml

from dossier.label import LabelStore
import kvlayer
import yakonfig

from trec_dd.harness.run import Harness
from trec_dd.harness.truth_data import parse_truth_data
from trec_dd.system.ambassador_cli import HarnessAmbassadorCLI

logger = logging.getLogger(__name__)

class RandomSystem(object):
    '''A ranking system that returns a random document id.
    '''

    def __init__(self, doc_store):
        self.doc_store = doc_store

    def search(self, query, page_number):
        '''Select 5 random documents.
        '''
        if page_number >= 5: return []
        doc_ids = list(self.doc_store.scan_ids(query))
        rand_docs = random.sample(doc_ids, min(len(doc_ids), 5))
        confidences = [str(int(1000 * random.random())) for _ in xrange(5)]
        results = list(chain(*zip(rand_docs, confidences)))
        assert len(results) % 2 == 0, results
        return results

    def process_feedback(self, feedback):
        '''Ignore the feedback from the harness.
        '''
        pass


class StubDocumentStore(object):
    '''Trivial document store.

    Allows for one to just iterate over a predefined set
    of document ids stored in memory.
    '''

    def __init__(self, topic_id_to_doc_ids):
        self.topic_id_to_doc_ids = topic_id_to_doc_ids

    def scan_ids(self, topic_id):
        '''Just iterate over doc ids in memory.
        '''
        for doc_id in self.topic_id_to_doc_ids[topic_id]:
            yield doc_id

def make_doc_store(label_store):

    all_topics = set()
    for label in label_store.everything():
        all_topics.add(str(label.meta['topic_id']))

    # build a silly document store. This store will just have
    # documents corresponding to the topic ids specified within
    # the topic sequence.
    topic_id_to_doc_ids = defaultdict(list)
    for label in label_store.everything():
        if label.content_id1 in all_topics:
            doc_id = label.content_id2
            if not doc_id.strip(): 
                logger.warn('skipping bogus document identifer: %r' % doc_id)
                continue
            topic_id = label.content_id1
            query = label.meta['topic_name']
            topic_id_to_doc_ids[query].append(doc_id)
        elif label.content_id2 in all_topics:
            doc_id = label.content_id1
            if not doc_id.strip(): 
                logger.warn('skipping bogus document identifer: %r' % doc_id)
                continue
            topic_id = label.content_id2
            query = label.meta['topic_name']
            topic_id_to_doc_ids[query].append(doc_id)
    doc_store = StubDocumentStore(topic_id_to_doc_ids)
    return doc_store

def main():
    '''Run the random recommender system on a sequence of topics.
    '''
    description = ('A baseline recommender system that uses the truth data to'
                   ' create output that has perfect recall and would also have'
                   ' perfect precision if you ignore subtopic diversity/novelty.'
                   ' This generates output directly from the truth data and'
                   ' randomly shuffles the truth data per topic, so that'
                   ' the ordering of passages does not attempt to optimize any'
                   ' particular quality metric.')
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--overwrite', action='store_true')
    args = yakonfig.parse_args(parser, [yakonfig])

    logging.basicConfig(level=logging.DEBUG)

    config = yakonfig.get_global_config('harness')
    batch_size = config.get('batch_size', 5)
    run_file_path = config['run_file_path']
    if os.path.exists(run_file_path):
        if args.overwrite:
            os.remove(run_file_path)
        else:
            sys.exit('%r already exists' % run_file_path)

    kvl_config = {'storage_type': 'local',
                  'namespace': 'test',
                  'app_name': 'test'}
    kvl = kvlayer.client(kvl_config)
    label_store = LabelStore(kvl)

    parse_truth_data(label_store, config['truth_data_path'])

    # Set up the system
    doc_store = make_doc_store(label_store)
    system = RandomSystem(doc_store)
    ambassador = HarnessAmbassadorCLI(system, args.config, batch_size)
    ambassador.run()


if __name__ == '__main__':
    main()



'''trec_dd.scorer.cube_test provides the
   cube test for each subtopic and then takes their mean.

   http://dl.acm.org/citation.cfm?id=2523648
   https://github.com/trec-dd/trec-dd-metrics/tree/master/cube-test

   This requires graded relevance.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''
from __future__ import division
import sys
from operator import attrgetter
from collections import defaultdict

#from numpy import mean

from trec_dd.utils import get_all_subtopics, get_best_subtopics

def mean(l):
    s = sum(l)
    return s / len(l)

def cube_test(run, label_store):
    '''
    iterate over topics and do cube test
    '''

    scores_by_topic = dict()

    ## score for each topic
    for topic_id, results in run['results'].items():
        
        ## get all subtopics for the topic
        subtopic_ids = list(set(get_all_subtopics(label_store, topic_id)))

        scores_by_topic[topic_id] = score_topic(topic_id, subtopic_ids, results)


       

    ## macro average over all the topics
    macro_avg = mean(scores_by_topic.values())

    run['scores']['cube_test'] = \
        {'scores_by_topic': scores_by_topic, 'macro_average': macro_avg}


def score_topic(topic_id, subtopic_ids, results):
    '''
    Runs the cube-test on a query topic.

    Cube-test requires computing the Gain of the results
    and the Time it takes to process them. The cube-test
    score is a measure of gain per time.

    `topic_id' is the current topic query to be scored
    `subtopic_ids' is a list of all the subtopics
    `results' is a dict containing results of the query,
              a ranked list
    '''

    ## this is where we would put relative importance of 
    ## the subtopics. at the moment, it's uniform
    theta = dict()
    for subtopic in subtopic_ids:
        theta[subtopic] = 1/len(subtopic_ids)

    ## this parameter sets the discount factor
    ## for topics already seen
    gamma = 0.5

    ## this dictionary allows us to compute the discount
    nrel = defaultdict(int)


    ## need to track when a subtopic overflows top of cube
    ## (i.e. when we fill up that subtopic)
    ## we assume the cube has unit height
    ## (really doesn't matter, is a normalization issue)
    tot_sub_rel = defaultdict(float)


    ## document length, in words
    ## since we don't collect, assuming default
    doc_len = 100

    time = 0
    score = 0

    for idx, result in enumerate(results):
        assert idx == result['rank'] - 1

        ## relevance of the document to the topic
        rel_to_topic = result['on_topic']

        gain_i = 0
        for subtopic, conf in get_best_subtopics(result['subtopics']):
            
            ## compute gain
            rel = (2**conf - 1)/2**4 ## magic formula for relevance
            Gamma = gamma**nrel[subtopic]
            gain_i += Gamma*theta[subtopic]*rel*(tot_sub_rel[subtopic] < 1)

            ## update parameters
            nrel[subtopic] += 1
            tot_sub_rel[subtopic] += rel


        ## compute the time for this doc
        if rel_to_topic:
            r_i = 0.64
        else:
            r_i = 0.39
        time += 4.4 + r_i*(0.018*doc_len +7.8)

        ## compute the cube-test score
        score += (1/len(results))*(gain_i/time)

        ## Note on above:
        ## for an infinite stream, this score goes to 0
        ## rather than converge to a constant
        ## I feel like it should not have the 1/len(results)
        ## out front, or maybe use marginal rather than cumulative time

    return score













'''trec_dd.utils provides tools for interacting with labels in TREC DD

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

'''
from collections import defaultdict

def get_all_subtopics(label_store, topic_id):
    labels = label_store.directly_connected(topic_id)

    def subtopic_from_label(label):
        subtopic_id = label.subtopic_for(topic_id)
        return subtopic_id

        ## oops, I thought we needed this for the special format used
        ## in TREC DD label table...
        offset, text = subtopic_id.split('|')
        text = topic_id_to_query(text)
        subtopic = {
            'subtopic_id': label.subtopic_for(topic),
            'offset': offset,
            'text': text,
            'rating': label.rating
        }
        return subtopic

    subtopics = map(subtopic_from_label, labels)
    return subtopics


def get_best_subtopics(subtopic_pairs):
    '''Return the instance of each subtopic with the highest rating.
    '''
    sid_to_data = defaultdict(list)
    for subtopic, conf in subtopic_pairs:
        sid_to_data[subtopic].append((subtopic, conf))
    subtopics = []
    for sid, data in sid_to_data.iteritems():
        best = max(data, key=lambda d: d[1])
        subtopics.append(best)
    return subtopics
